from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")


print("After user connect him account you will receive in your webhook a code for authenticate")

code = 'ec2c1ee4d1f40f4db60ea30f4ea4769d'
token_resp = fintecture.AIS.oauth(code=code)

fintecture.access_token = token_resp['access_token']

print("Success: %r" % (token_resp))



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from fintecture.api_resources.abstract import CreateableAPIResource
from fintecture.api_resources.abstract import DeletableAPIResource
from fintecture.api_resources.abstract import ListableAPIResource
from fintecture.api_resources.abstract import UpdateableAPIResource


class AccountHolder(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "accountholder"

    @classmethod
    def search_by_customer(cls, customer_id, **params):
        return cls._static_request(
            "get",
            "/ais/v1/customer/{}/accountholders".format(
                customer_id
            ),
            params=params,
        )



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from fintecture.fintecture_object import FintectureObject


class UsageRecordSummary(FintectureObject):
    OBJECT_NAME = "usage_record_summary"



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")


print("Searching URL to connect with AIS...")

resp = fintecture.AIS.connect(redirect_uri="https://domain.com", state="1234")

print("Redirecting user to {} ...".format(
    resp['meta']['url']
))
print("Success: %r" % (resp))



from __future__ import absolute_import, division, print_function

import functools
import hmac
import io
import logging
import sys
import os
import re
import time
import threading

import fintecture
from fintecture import six
from fintecture.six.moves.urllib.parse import parse_qsl, quote_plus

FINTECTURE_LOG = os.environ.get("FINTECTURE_LOG")

logger = logging.getLogger("fintecture")

__all__ = [
    "io",
    "parse_qsl",
    "utf8",
    "log_info",
    "log_debug",
    "dashboard_link",
    "logfmt",
]


def utf8(value):
    if six.PY2 and isinstance(value, six.text_type):
        return value.encode("utf-8")
    else:
        return value


def is_appengine_dev():
    return "APPENGINE_RUNTIME" in os.environ and "Dev" in os.environ.get(
        "SERVER_SOFTWARE", ""
    )


def _console_log_level():
    if fintecture.log in ["debug", "info"]:
        return fintecture.log
    elif FINTECTURE_LOG in ["debug", "info"]:
        return FINTECTURE_LOG
    else:
        return None


def log_debug(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() == "debug":
        print(msg, file=sys.stderr)
    logger.debug(msg)


def log_info(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() in ["debug", "info"]:
        print(msg, file=sys.stderr)
    logger.info(msg)


def _test_or_live_environment():
    if fintecture.api_key is None:
        return
    match = re.match(r"sk_(live|test)_", fintecture.api_key)
    if match is None:
        return
    return match.groups()[0]


def dashboard_link(request_id):
    return "https://dashboard.fintecture.com/{env}/logs/{reqid}".format(
        env=_test_or_live_environment() or "test", reqid=request_id
    )


def logfmt(props):
    def fmt(key, val):
        # Handle case where val is a bytes or bytesarray
        if six.PY3 and hasattr(val, "decode"):
            val = val.decode("utf-8")
        # Check if val is already a string to avoid re-encoding into
        # ascii. Since the code is sent through 2to3, we can't just
        # use unicode(val, encoding='utf8') since it will be
        # translated incorrectly.
        if not isinstance(val, six.string_types):
            val = six.text_type(val)
        if re.search(r"\s", val):
            val = repr(val)
        # key should already be a string
        if re.search(r"\s", key):
            key = repr(key)
        return u"{key}={val}".format(key=key, val=val)

    return u" ".join([fmt(key, val) for key, val in sorted(props.items())])


# Borrowed from Django's source code
if hasattr(hmac, "compare_digest"):
    # Prefer the stdlib implementation, when available.
    def secure_compare(val1, val2):
        return hmac.compare_digest(utf8(val1), utf8(val2))

else:

    def secure_compare(val1, val2):
        """
        Returns True if the two strings are equal, False otherwise.
        The time taken is independent of the number of characters that match.
        For the sake of simplicity, this function executes in constant time
        only when the two strings have the same length. It short-circuits when
        they have different lengths.
        """
        val1, val2 = utf8(val1), utf8(val2)
        if len(val1) != len(val2):
            return False
        result = 0
        if six.PY3 and isinstance(val1, bytes) and isinstance(val2, bytes):
            for x, y in zip(val1, val2):
                result |= x ^ y
        else:
            for x, y in zip(val1, val2):
                result |= ord(x) ^ ord(y)
        return result == 0


def get_object_classes():
    # This is here to avoid a circular dependency
    from fintecture.object_classes import OBJECT_CLASSES

    return OBJECT_CLASSES


def convert_to_fintecture_object(
    resp, app_id=None, fintecture_version=None, fintecture_account=None, params=None
):
    # If we get a FintectureResponse, we'll want to return a
    # FintectureObject with the last_response field filled out with
    # the raw API response information
    fintecture_response = None

    if isinstance(resp, fintecture.fintecture_response.FintectureResponse):
        fintecture_response = resp
        resp = fintecture_response.data

    if isinstance(resp, list):
        return [
            convert_to_fintecture_object(
                i, app_id, fintecture_version, fintecture_account
            )
            for i in resp
        ]
    elif isinstance(resp, dict) and not isinstance(
        resp, fintecture.fintecture_object.FintectureObject
    ):
        resp = resp.copy()
        klass_name = resp.get("object")
        if isinstance(klass_name, six.string_types):
            klass = get_object_classes().get(
                klass_name, fintecture.fintecture_object.FintectureObject
            )
        else:
            klass = fintecture.fintecture_object.FintectureObject

        obj = klass.construct_from(
            resp,
            app_id,
            fintecture_version=fintecture_version,
            fintecture_account=fintecture_account,
            last_response=fintecture_response,
        )

        # We only need to update _retrieve_params when special params were
        # actually passed. Otherwise, leave it as is as the list / search result
        # constructors will instantiate their own params.
        if (
            params is not None
            and hasattr(obj, "object")
            and ((obj.object == "list") or (obj.object == "search_result"))
        ):
            obj._retrieve_params = params

        return obj
    else:
        return resp


def convert_to_dict(obj):
    """Converts a FintectureObject back to a regular dict.

    Nested FintectureObjects are also converted back to regular dicts.

    :param obj: The FintectureObject to convert.

    :returns: The FintectureObject as a dict.
    """
    if isinstance(obj, list):
        return [convert_to_dict(i) for i in obj]
    # This works by virtue of the fact that FintectureObjects _are_ dicts. The dict
    # comprehension returns a regular dict and recursively applies the
    # conversion to each value.
    elif isinstance(obj, dict):
        return {k: convert_to_dict(v) for k, v in six.iteritems(obj)}
    else:
        return obj


def populate_headers(idempotency_key):
    if idempotency_key is not None:
        return {"Idempotency-Key": idempotency_key}
    return None


def read_special_variable(params, key_name, default_value):
    value = default_value
    params_value = None

    if params is not None and key_name in params:
        params_value = params[key_name]
        del params[key_name]

    if value is None:
        value = params_value

    return value


def merge_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


def sanitize_id(id):
    utf8id = utf8(id)
    quotedId = quote_plus(utf8id)
    return quotedId


def set_interval(func, sec):
    def func_wrapper():
        set_interval(func, sec)
        func()
    t = threading.Timer(sec, func_wrapper)
    t.start()
    return t


class Interval:
    def __init__(self, interval, action):
        self.interval = interval
        self.action = action
        self.stopEvent = threading.Event()
        thread = threading.Thread(target=self._set_interval)
        thread.start()

    def _set_interval(self):
        next_time = time.time() + self.interval
        while not self.stopEvent.wait(next_time-time.time()):
            next_time += self.interval
            self.action()

    def cancel(self) :
        self.stopEvent.set()


class class_method_variant(object):

    def __init__(self, class_method_name):
        self.class_method_name = class_method_name

    def __call__(self, method):
        self.method = method
        return self

    def __get__(self, obj, objtype=None):
        @functools.wraps(self.method)
        def _wrapper(*args, **kwargs):
            if obj is not None:
                # Method was called as an instance method, e.g.
                # instance.method(...)
                return self.method(obj, *args, **kwargs)
            elif len(args) > 0 and isinstance(args[0], objtype):
                # Method was called as a class method with the instance as the
                # first argument, e.g. Class.method(instance, ...) which in
                # Python is the same thing as calling an instance method
                return self.method(args[0], *args[1:], **kwargs)
            else:
                # Method was called as a class method, e.g. Class.method(...)
                class_method = getattr(objtype, self.class_method_name)
                return class_method(*args, **kwargs)

        return _wrapper



from __future__ import absolute_import, division, print_function

import base64
import fintecture

from fintecture import api_requestor, error
from fintecture.six.moves.urllib.parse import urlencode


class OAuth(object):

    @staticmethod
    def _set_app_params(params):
        if "app_id" in params and "app_secret" in params:
            return

        from fintecture import app_id, app_secret

        no_params = False
        if app_id:
            params["app_id"] = app_id
        else:
            no_params = True

        if app_secret:
            params["app_secret"] = app_secret
        else:
            no_params = True

        if no_params:
            raise error.AuthenticationError(
                "No app_id and/or app_secret provided. (HINT: set your app_id and app_secret using "
                '"fintecture.app_id = <APP-ID>" and "fintecture.app_secret = <APP-SECRET>"). '
                "You can find your application values in your Fintecture developer console at "
                "https://console.fintecture.com/developers, after registering your account as a platform. "
                "See https://docs.fintecture.com/ for details, or email contact@fintecture.com "
                "if you have any questions."
            )

    @staticmethod
    def ais_token(app_id=None, app_secret=None, **params):
        if app_id:
            params['app_id'] = app_id
        if app_secret:
            params['app_secret'] = app_secret
        OAuth._set_app_params(params)
        return fintecture.AIS.oauth(params)

    @staticmethod
    def pis_token(app_id=None, app_secret=None, **params):
        if app_id:
            params['app_id'] = app_id
        if app_secret:
            params['app_secret'] = app_secret
        OAuth._set_app_params(params)
        return fintecture.PIS.oauth(params)

    @staticmethod
    def refresh_token(app_id=None, app_secret=None, **params):
        if not params.get('refresh_token', False):
            raise error.InvalidRequestError(
                "refresh_token parameter is required for refresh token"
            )

        if app_id:
            params['app_id'] = app_id
        if app_secret:
            params['app_secret'] = app_secret

        OAuth._set_app_params(params)

        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': "Basic {}".format(
                base64.b64encode("{}:{}".format(
                    params["app_id"],
                    params['app_secret']
                ).encode('utf-8')).decode('utf-8')
            )
        }

        params.update({
            'grant_type': 'refresh_token',
        })
        requestor = api_requestor.APIRequestor(
            app_id=params['app_id'],
            app_secret=params['app_secret'],
        )

        response, _ = requestor.request(
            "post", "/oauth/refreshtoken", params, headers
        )
        return response.data



# -*- coding: utf-8 -*-

import json
import uuid
import base64

from cryptography.hazmat.primitives.asymmetric import dsa, rsa, padding
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.exceptions import UnsupportedAlgorithm

from hashlib import sha256


def generate_uuid():
    return str(uuid.uuid4()).replace('-', '')


def generate_uuidv4():
    return str(uuid.uuid4())


def hash_base64(plain_text):
    m = sha256()
    m.update(plain_text.encode('utf-8'))
    b_digest = m.digest()
    b_b64 = base64.b64encode(b_digest)
    plain_b64_hash = b_b64.decode('ascii')
    return plain_b64_hash


def create_signature_header(headers, app_id, private_key, signed_headers):
    signing_string = build_signing_string(headers, signed_headers)
    header_string = build_header_string(headers, signed_headers)

    signature = sign_payload(signing_string, private_key)
    return (
            'keyId="' + app_id +
            '",algorithm="rsa-sha256",headers="' + header_string +
            '",signature="' + signature + '"'
    )


def build_signing_string(headers, signed_headers):
    signing_string = ''

    for param in signed_headers:
        if param in headers:
            p = param.lower()
            signing_string = signing_string + '\n' if signing_string else signing_string
            signing_string = signing_string + p + ': ' + headers[param]

    return signing_string


def build_header_string(headers, signed_headers):
    header_string = ''

    for param in signed_headers:
        if param in headers:
            p = param.lower()
            header_string = header_string + ' ' + p if header_string else p

    return header_string


def sign_payload(payload, private_key, algorithm=None):
    if isinstance(payload, dict):
        payload = json.dumps(payload, separators=(',', ':'))

    bytes_payload = str.encode(payload)

    if not algorithm or algorithm == 'rsa-sha256':
        try:
            bytes_private_key = str.encode(private_key)
            private_key = load_pem_private_key(bytes_private_key, password=None)

            signature = private_key.sign(bytes_payload, padding.PKCS1v15(), hashes.SHA256())

            base64_bytes = base64.b64encode(signature)
            plain_b64_signature = base64_bytes.decode('ascii')

            return plain_b64_signature

        except ValueError as error:
            raise ValueError(
                'the PEM data could not be decrypted or '
                'if its structure could not be decoded successfully'
            )
        except TypeError as error:
            raise ValueError(
                'password was given and the private key was not encrypted; '
                'or the key was encrypted but no password was supplied'
            )
        except UnsupportedAlgorithm as error:
            raise ValueError(
                'the serialized key type is not supported by the OpenSSL version cryptography is using'
            )

    raise ValueError('invalid signature algorithm')



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from fintecture import util
from fintecture.api_resources.abstract import CreateableAPIResource
from fintecture.api_resources.abstract import ListableAPIResource
from fintecture.api_resources.abstract import UpdateableAPIResource


class VerificationSession(
    CreateableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "identity.verification_session"

    @classmethod
    def _cls_cancel(
        cls,
        session,
        api_key=None,
        fintecture_version=None,
        fintecture_account=None,
        **params
    ):
        return cls._static_request(
            "post",
            "/v1/identity/verification_sessions/{session}/cancel".format(
                session=util.sanitize_id(session)
            ),
            api_key=api_key,
            fintecture_version=fintecture_version,
            fintecture_account=fintecture_account,
            params=params,
        )

    @util.class_method_variant("_cls_cancel")
    def cancel(self, idempotency_key=None, **params):
        return self._request(
            "post",
            "/v1/identity/verification_sessions/{session}/cancel".format(
                session=util.sanitize_id(self.get("id"))
            ),
            idempotency_key=idempotency_key,
            params=params,
        )

    @classmethod
    def _cls_redact(
        cls,
        session,
        api_key=None,
        fintecture_version=None,
        fintecture_account=None,
        **params
    ):
        return cls._static_request(
            "post",
            "/v1/identity/verification_sessions/{session}/redact".format(
                session=util.sanitize_id(session)
            ),
            api_key=api_key,
            fintecture_version=fintecture_version,
            fintecture_account=fintecture_account,
            params=params,
        )

    @util.class_method_variant("_cls_redact")
    def redact(self, idempotency_key=None, **params):
        return self._request(
            "post",
            "/v1/identity/verification_sessions/{session}/redact".format(
                session=util.sanitize_id(self.get("id"))
            ),
            idempotency_key=idempotency_key,
            params=params,
        )



from __future__ import absolute_import, division, print_function

from fintecture.api_resources.abstract import SingletonAPIResource


class Application(SingletonAPIResource):
    OBJECT_NAME = "application"

    @classmethod
    def class_url(cls):
        return "/res/v1/applications"



from __future__ import absolute_import, division, print_function

from fintecture.api_resources.abstract.api_resource import APIResource


class SearchableAPIResource(APIResource):
    @classmethod
    def _search(
        cls,
        search_url,
        app_id=None,
        fintecture_version=None,
        fintecture_account=None,
        **params
    ):
        return cls._static_request(
            "get",
            search_url,
            app_id=app_id,
            fintecture_version=fintecture_version,
            fintecture_account=fintecture_account,
            params=params,
        )



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")
fintecture.private_key = os.environ.get("FINTECTURE_PRIVATE_KEY")


print("Searching URL to connect with PIS...")

resp = fintecture.PIS.oauth()

print("Success: %r" % (resp))

fintecture.access_token = resp['access_token']

print("Connecting with PIS...")

resp_connect = fintecture.PIS.connect(
    state="1234",
    meta={
        'psu_name': 'M. John Doe',
        'psu_email': 'john@doe.com',
        'psu_phone': '0601020304',
        'psu_ip': '127.0.0.1',
        'psu_address': {
            'street': '5 Void Street',
            'complement': 'RDC',
            'zip': '12345',
            'city': 'Gotham',
            'country': 'FR',
        },
    },
    data={
        'type': 'SEPA',
        'attributes': {
            'amount': '550.60',
            'currency': 'EUR',
            'communication': 'Order 15654'
        }
    }
)

print("Success: %r" % (resp_connect))



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")
fintecture.private_key = os.environ.get("FINTECTURE_PRIVATE_KEY")
fintecture.access_token = os.environ.get("FINTECTURE_ACCESS_TOKEN")


print("Retrieving all payments...")

if fintecture.access_token is None:
    print("Requesting an access token...")

    resp = fintecture.PIS.oauth()

    print("Success: %r" % (resp))

    fintecture.access_token = resp['access_token']
else:
    print("Using access token: %s" % fintecture.access_token)

print("Now retrieves payments...")

resp_payments = fintecture.Payment.search()

print("Success: %r" % (resp_payments))

print("Select one session identifier to filter payments by session...")
session_id = resp_payments.get('data', [])[0].get('id', False)

payment = fintecture.Payment.retrieve(session_id)

print("Success: %r" % (payment))


print("Canceling payment...")

resp_payment_cancellation = payment.update(status='payment_cancelled')

print("Success: %r" % (resp_payment_cancellation))



from __future__ import absolute_import, division, print_function

import os

import fintecture

fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")

print("Attempting charge...")

proxy = {
    "http": "http://<user>:<pass>@<proxy>:<port>",
    "https": "http://<user>:<pass>@<proxy>:<port>",
}

clients = (
    fintecture.http_client.RequestsClient(
        verify_ssl_certs=fintecture.verify_ssl_certs, proxy=proxy
    ),
    fintecture.http_client.PycurlClient(
        verify_ssl_certs=fintecture.verify_ssl_certs, proxy=proxy
    ),
    fintecture.http_client.Urllib2Client(
        verify_ssl_certs=fintecture.verify_ssl_certs, proxy=proxy
    ),
)

for c in clients:
    fintecture.default_http_client = c
    resp = fintecture.PIS.connect(
        state="1234",
        meta={
            'psu_name': 'M. John Doe',
            'psu_email': 'john@doe.com',
            'psu_phone': '0601020304',
            'psu_ip': '127.0.0.1',
            'psu_address': {
                'street': '5 Void Street',
                'complement': 'RDC',
                'zip': '12345',
                'city': 'Gotham',
                'country': 'FR',
            },
        },
        data={
            'type': 'SEPA',
            'attributes': {
                'amount': '550.60',
                'currency': 'EUR',
                'communication': 'Order 15654'
            }
        }
    )
    print("Success: %s, %r" % (c.name, resp))



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")


try:
    print("Searching URL to connect with PIS...")

    resp = fintecture.PIS.oauth()

    print("Success: %r" % (resp))

    fintecture.access_token = resp['access_token']

    print("Initiating with PIS...")

    provider = "baskes2b"
    redirect_uri = "https://domain.com"

    resp_connect = fintecture.PIS.initiate(
        provider,
        redirect_uri,
        state="1234",
        meta={
            'psu_name': 'M. John Doe',
            'psu_email': 'john@doe.com',
            'psu_phone': '0601020304',
            'psu_ip': '127.0.0.1',
            'psu_address': {
                'street': '5 Void Street',
                'complement': 'RDC',
                'zip': '12345',
                'city': 'Gotham',
                'country': 'FR',
            },
        },
        data={
            'type': 'SEPA',
            'attributes': {
                'amount': '550.60',
                'currency': 'EUR',
                'communication': 'Order 15654'
            }
        }
    )

    print("Success: %r" % (resp_connect))

except Exception as e:
    print("ERROR received: %r" % e)



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from fintecture.api_resources.abstract import CreateableAPIResource
from fintecture.api_resources.abstract import DeletableAPIResource
from fintecture.api_resources.abstract import ListableAPIResource
from fintecture.api_resources.abstract import SearchableAPIResource
from fintecture.api_resources.abstract import UpdateableAPIResource
from fintecture.api_resources.account import Account
from fintecture.api_resources.account_holder import AccountHolder


class Customer(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    SearchableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "customer"

    @classmethod
    def get_accounts(cls, customer_id, **params):
        return Account.search_by_customer(customer_id)

    @classmethod
    def get_account_holders(cls, customer_id, **params):
        return AccountHolder.search_by_customer(customer_id)

    @classmethod
    def get_account_transactions(cls, customer_id, account_id, **params):
        return Account.search_transactions_by_customer_account(
            customer_id, account_id
        )

    @classmethod
    def delete(cls, customer_id, **params):
        return Account.delete(customer_id, params)

    @classmethod
    def class_url(cls):
        return "/ais/v1/customer"



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from fintecture.api_resources.abstract import ListableAPIResource
from fintecture.api_resources.abstract import SearchableAPIResource


class Provider(
    ListableAPIResource,
    SearchableAPIResource,
):
    OBJECT_NAME = "provider"

    @classmethod
    def search(cls, *args, **kwargs):
        return cls._search(search_url="/res/v1/providers", *args, **kwargs)

    @classmethod
    def search_auto_paging_iter(cls, *args, **kwargs):
        return cls.search(*args, **kwargs).auto_paging_iter()



from __future__ import absolute_import, division, print_function

import os

import fintecture

from fintecture.constants import DECOUPLED_MODEL_TYPE
from fintecture.six.moves.urllib.parse import urlparse, urlsplit


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")
fintecture.private_key = os.environ.get("FINTECTURE_PRIVATE_KEY")


try:
    print("Do operations using a decoupled model with AIS...")

    # 1. select a Bank (provider)
    # provider_id = "bschesmm"  # Banco Santander
    # provider_id = "bsabesbb"  # Banco Sabadell
    # provider_id = "cmcifrpp"  # CIC
    provider_id = "handse"  # Handelsbanken | This is an only one provider that support decoupled model
    redirect_uri = 'https://webhook.site/5940c656-7fd3-40a0-ad41-9cfba1e986f3'


    # 2. get the provider's decoupled auth polling URL
    resp_authorize = fintecture.AIS.authorize(
        provider_id=provider_id,
        redirect_uri=redirect_uri,
        state="1234",
        model=DECOUPLED_MODEL_TYPE,
        psu_id='9090901',
        psu_ip_address='92.168.0.12',
    )

    provider_auth_url = resp_authorize['url']
    print("Redirecting user to {} ...".format(
        provider_auth_url
    ))
    # provider_auth_polling_id = resp_authorize['polling_id']
    # print("Polling identifier is {} ...".format(
    #     provider_auth_polling_id
    # ))

    parsed_url = urlparse(provider_auth_url)
    path = parsed_url.path
    provider_auth_polling_id = path.rsplit('/', 1)[-1]
    # provider_auth_polling_id = 'aee32ed6-73fc-4154-9f18-149169b0bc6f'
    print("polling auth identifier: %s\n" % (provider_auth_polling_id))
    print("Success: %r\n" % (resp_authorize))


    # 3. keep polling the provider's decoupled auth polling URL until authentication either COMPLETED or FAILED
    customer_id = None
    code = None

    # for now, we use a simple request for check status of decoupled authentication,
    # but you must do a keep polling requests for check status until it is COMPLETED or FAILED
    resp_decoupled = fintecture.AIS.decoupled(
        provider_id=provider_id,
        polling_id=provider_auth_polling_id,
        redirect_uri=redirect_uri,
        state="1234",
        psu_id='9090901',
        psu_ip_address='92.168.0.12',
    )
    print("Decoupled response: %r\n" % (resp_decoupled))

    decoupled_auth_status = resp_decoupled['status']
    if decoupled_auth_status == 'COMPLETED':
        # IMPORTANT: we are replacing expected property 'customer_id' and received 'connection_id'
        # customer_id = resp_decoupled['customer_id']
        customer_id = resp_decoupled['connection_id']
        code = resp_decoupled['code']
    elif decoupled_auth_status == 'FAILED':
        raise ValueError("Decoupled authentication fails for an unknown reason. Try later!")
    else:
        raise ValueError("Decoupled authentication is yet pending. Keep polling for check status!")


    # 4. authenticate your app to Fintecture and get your "access_token" and "refresh_token"
    resp_token = fintecture.AIS.oauth(code=code)

    fintecture.access_token = resp_token['access_token']
    refresh_token = resp_token['refresh_token']
    expires_in = int(resp_token['expires_in'])
    print("Access token is: %s" % (fintecture.access_token))
    print("Refresh token is: %s" % (refresh_token))

    print("Success: %r\n" % (resp_token))


    # 5. request any AIS API
    accounts = fintecture.Account.search_by_customer(customer_id)

    print("accounts: %r" % (accounts))


    # 6. request a refresh token
    refresh_token_resp = fintecture.OAuth.refresh_token(
        refresh_token=refresh_token
    )
    fintecture.access_token = refresh_token_resp['access_token']
    print("Access token was refreshed with the new value: %s" % (fintecture.access_token))

    print("Success: %r\n" % (refresh_token_resp))

except Exception as e:
    print("ERROR received: %r" % e)



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")


print("Searching providers...")

resp = fintecture.Provider.search()

print("Success: %r" % (resp))



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")
fintecture.private_key = os.environ.get("FINTECTURE_PRIVATE_KEY")
fintecture.access_token = os.environ.get("FINTECTURE_ACCESS_TOKEN")


print("Retrieving all outgoing payments from your Local Acquiring account to your own bank account. ...")

if fintecture.access_token is None:
    print("Requesting an access token...")

    resp = fintecture.PIS.oauth()

    print("Success: %r" % (resp))

    fintecture.access_token = resp['access_token']
else:
    print("Using access token: %s" % fintecture.access_token)

print("Now retrieves payments...")

resp_settlements = fintecture.PIS.settlements()

print("Success: %r" % (resp_settlements))



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

# flake8: noqa

from fintecture.api_resources.error_object import ErrorObject, OAuthErrorObject
from fintecture.api_resources.list_object import ListObject
from fintecture.api_resources.search_result_object import SearchResultObject

from fintecture.api_resources import identity
from fintecture.api_resources import test_helpers

from fintecture.api_resources.ais import AIS
from fintecture.api_resources.application import Application
from fintecture.api_resources.account import Account
from fintecture.api_resources.account_holder import AccountHolder
from fintecture.api_resources.capability import Capability
from fintecture.api_resources.charge import Charge
from fintecture.api_resources.customer import Customer

from fintecture.api_resources.ephemeral_key import EphemeralKey
from fintecture.api_resources.payout import Payout
from fintecture.api_resources.pis import PIS
from fintecture.api_resources.payment import Payment
from fintecture.api_resources.provider import Provider
from fintecture.api_resources.refund import Refund
from fintecture.api_resources.test_account import TestAccount
from fintecture.api_resources.token import Token
from fintecture.api_resources.usage_record import UsageRecord
from fintecture.api_resources.usage_record_summary import UsageRecordSummary
from fintecture.api_resources.webhook_endpoint import WebhookEndpoint



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")
fintecture.access_token = os.environ.get("FINTECTURE_ACCESS_TOKEN")


print("Searching customer accounts...")

customer_id = 'f1923614953cb0e1f1da8b5c016fd5a7'

resp = fintecture.Customer.get_accounts(customer_id)

print("Success: %r" % (resp))

print("Searching customer transactions of the first result...")

transactions_resp = fintecture.Customer.get_account_transactions(customer_id, resp['data'][0]['id'])

print("Success: %r" % (transactions_resp))

print("Searching customer account holders...")

accountholders_resp = fintecture.Customer.get_account_holders(customer_id)

print("Success: %r" % (accountholders_resp))



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from fintecture import util
from fintecture import error

from fintecture.api_resources.abstract import SearchableAPIResource
from fintecture.api_resources.abstract import CreateableAPIResource
from fintecture.api_resources.abstract import DeletableAPIResource
from fintecture.api_resources.abstract import ListableAPIResource
from fintecture.api_resources.abstract import UpdateableAPIResource


class Payment(
    SearchableAPIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "payment"

    @classmethod
    def search(cls, *args, **kwargs):
        return cls._search(search_url="/pis/v2/payments", *args, **kwargs)

    @classmethod
    def search_auto_paging_iter(cls, *args, **kwargs):
        return cls.search(*args, **kwargs).auto_paging_iter()

    @classmethod
    def connect(cls, **params):
        if not params.get('state', False):
            raise error.InvalidRequestError(
                "state: A state parameter which will be provided back on redirection."
            )
        state = params.get('state')
        del params['state']
        params.update({
            'headers': {
                'Content-Type': 'application/json'
            }
        })
        return cls._static_request(
            "post",
            "/pis/v2/connect?state={}".format(state),
            params=params,
        )

    @classmethod
    def initiate(cls, provider_id, redirect_uri, **params):
        if not params.get('state', False):
            raise error.InvalidRequestError(
                "state: A state parameter which will be provided back on redirection."
            )
        state = params.get('state')
        del params['state']
        params.update({
            'headers': {
                'Content-Type': 'application/json'
            }
        })
        return cls._static_request(
            "post",
            "/pis/v2/provider/{provider}/initiate?redirect_uri={url}&state={state}".format(
                provider=provider_id,
                state=state,
                url=redirect_uri
            ),
            params=params,
        )

    def refund(self, **params):
        params.update({
            "meta": {
                "session_id": self.get('data', {}).get('id', {})
            },
            'headers': {
                'Content-Type': 'application/json'
            }
        })
        return self._request(
            "post",
            "/pis/v2/refund",
            params=params,
        )

    def update(self, **params):
        session_id = self.get('data', {}).get('id', {})
        if not params.get('status', False):
            raise error.InvalidRequestError(
                "status: is a parameter for update payment attributes and only accepts 'payment_cancelled' value"
            )
        status = params.get('status')
        del params['status']

        params.update({
            "meta": {
                "status": status
            },
            'headers': {
                'Content-Type': 'application/json'
            }
        })
        return self._request(
            "patch",
            "/pis/v2/payments/{}".format(session_id),
            params=params,
        )

    @classmethod
    def class_url(cls):
        return "/pis/v2/payments"



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

import base64
import fintecture

from fintecture import error

from fintecture.api_resources.abstract import APIResource
from fintecture.api_resources.payment import Payment


class PIS(APIResource):
    OBJECT_NAME = "pis"

    @classmethod
    def connect(cls, **params):
        return Payment.connect(**params)

    # noinspection PyPackageRequirements
    @classmethod
    def initiate(cls, provider_id, redirect_uri, **params):
        return Payment.initiate(provider_id, redirect_uri, **params)

    @classmethod
    def oauth(cls, **params):
        params.update({
            'app_id': fintecture.app_id,
            'grant_type': 'client_credentials',
            'scope': 'PIS',
            'headers': {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': "Basic {}".format(
                    base64.b64encode("{}:{}".format(
                        fintecture.app_id,
                        fintecture.app_secret
                    ).encode('utf-8')).decode('utf-8')
                )
            }
        })
        return cls._static_request(
            "post",
            "/oauth/accesstoken",
            params=params,
        )

    @classmethod
    def initiate_refund(cls, **params):
        if not params.get('state', False):
            raise error.InvalidRequestError(
                "state: A state parameter which will be provided back on redirection."
            )
        state = params.get('state')
        del params['state']
        params.update({
            'headers': {
                'Content-Type': 'application/json'
            }
        })
        return cls._static_request(
            "post",
            "/pis/v2/refund?state={}".format(state),
            params=params,
        )

    @classmethod
    def request_to_pay(cls, redirect_uri, **params):
        language = params.get('language', '')
        if not language:
            raise error.InvalidRequestError(
                "language: its a required parameter that is a code of two letters; ex: fr"
            )
        if len(language) != 2:
            raise error.InvalidRequestError(
                "language parameter is of two characters of length; ex: fr"
            )
        params.update({
            'headers': {
                'Content-Type': 'application/json',
                'x-language': language,
            }
        })
        return cls._static_request(
            "post",
            "/pis/v2/request-to-pay?redirect_uri={url}".format(url=redirect_uri),
            params=params,
        )

    @classmethod
    def request_to_payout(cls, redirect_uri, **params):
        if not params.get('state', False):
            raise error.InvalidRequestError(
                "state: A state parameter which will be provided back on redirection."
            )
        state = params.get('state')
        del params['state']

        language = params.get('language', '')
        if not language:
            raise error.InvalidRequestError(
                "language: its a required parameter that is a code of two letters; ex: fr"
            )
        if len(language) != 2:
            raise error.InvalidRequestError(
                "language parameter is of two characters of length; ex: fr"
            )
        params.update({
            'headers': {
                'Content-Type': 'application/json',
                'x-language': language,
            }
        })
        return cls._static_request(
            "post",
            "/pis/v2/request-for-payout?redirect_uri={url}&state={state}".format(
                url=redirect_uri,
                state=state,
            ),
            params=params,
        )

    @classmethod
    def settlements(cls, **params):
        settlement_id = params.get('settlement_id', False)
        if settlement_id:
            del params['settlement_id']
            settlement_id = '/%s' % settlement_id
        else:
            settlement_id = ''

        params.update({
            'headers': {
                'Content-Type': 'application/json'
            }
        })
        return cls._static_request(
            "get",
            "/pis/v2/settlements{settlement}".format(settlement=settlement_id),
            params=params,
        )

    @classmethod
    def class_url(cls):
        return "/v1/pis"



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

import base64
import fintecture

from fintecture import error

from fintecture.api_resources.abstract import APIResource
from fintecture.constants import DECOUPLED_MODEL_TYPE


class AIS(APIResource):
    OBJECT_NAME = "ais"

    @classmethod
    def oauth(cls, **params):
        if not params.get('code', False):
            raise error.InvalidRequestError(
                "code parameter is required for authenticate with oAuth through AIS application"
            )

        params.update({
            'grant_type': 'authorization_code',
            'scope': 'AIS',
            'headers': {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': "Basic {}".format(
                    base64.b64encode("{}:{}".format(
                        fintecture.app_id,
                        fintecture.app_secret
                    ).encode('utf-8')).decode('utf-8')
                )
            }
        })
        return cls._static_request(
            "post",
            "/oauth/accesstoken",
            params=params,
        )

    @classmethod
    def connect(cls, **params):
        if not params.get('redirect_uri', False):
            raise error.InvalidRequestError(
                "redirect_uri: must correspond to one of the URLs provided when creating an application on the console."
            )
        if not params.get('state', False):
            raise error.InvalidRequestError(
                "state: an mandatory state parameter which will be provided back on redirection	."
            )
        return cls._static_request(
            "get",
            "/ais/v2/connect",
            params=params,
        )

    @classmethod
    def authorize(cls, provider_id, redirect_uri, **params):

        if not params.get('state', False):
            raise error.InvalidRequestError(
                "state: an mandatory state parameter which will be provided back on redirection	."
            )

        model = params.get('model', False)
        # access_token = params.get('access_token', False)

        headers = {}

        if model == DECOUPLED_MODEL_TYPE:
            psu_id = params.get('psu_id', False)
            psu_ip_address = params.get('psu_ip_address', False)
            if not psu_id or not psu_ip_address:
                raise error.InvalidRequestError(
                    "when model is 'decoupled' the 'x-psu-id' and 'x-psu-ip-address' parameters are required."
                )
            headers['x-psu-id'] = psu_id
            headers['x-psu-ip-address'] = psu_ip_address

        params.update({
            'response_type': 'code',
            'redirect_uri': redirect_uri,
            # 'access_token': access_token,
            'headers': headers
        })

        return cls._static_request(
            "get",
            "/ais/v1/provider/{provider}/authorize".format(provider=provider_id),
            params=params,
        )

    @classmethod
    def decoupled(cls, provider_id, polling_id, **params):
        headers = {}
        psu_id = params.get('psu_id', False)
        psu_ip_address = params.get('psu_ip_address', False)
        if not psu_id or not psu_ip_address:
            raise error.InvalidRequestError(
                "when model is 'decoupled' the 'x-psu-id' and 'x-psu-ip-address' parameters are required."
            )
        headers['x-psu-id'] = psu_id
        headers['x-psu-ip-address'] = psu_ip_address

        params.update({
            'response_type': 'code',
            'headers': headers
        })

        return cls._static_request(
            "get",
            "/ais/v1/provider/{provider}/authorize/decoupled/{polling}".format(
                provider=provider_id,
                polling=polling_id,
            ),
            params=params,
        )

    @classmethod
    def class_url(cls):
        return "/v1/ais"



from __future__ import absolute_import, division, print_function

import os

# Fintecture Python bindings
# API docs at https://docs.fintecture.com/
# Authors:
# Gabriel Urresta Romero <gabriel@melkart.io>
# Gilberto Cuba Ricardo <gilberto@melkart.io>

# API resources
from fintecture.environments import *  # noqa

# Configuration variables

app_id = None
app_secret = None
private_key = None
environment = None

access_token = None
api_key = None
client_id = None

production_api_base = "https://api.fintecture.com"
sandbox_api_base = "https://api-sandbox.fintecture.com"

env = DEFAULT_ENV

api_version = None
verify_ssl_certs = True
proxy = None
default_http_client = None
app_info = None
enable_telemetry = True
max_network_retries = 0
ca_bundle_path = os.path.join(
    os.path.dirname(__file__), "data", "ca-certificates.crt"
)

# Set to either 'debug' or 'info', controls console logging
log = None

# API resources
from fintecture.api_resources import *  # noqa

# OAuth
from fintecture.oauth import OAuth  # noqa

# Webhooks
from fintecture.webhook import Webhook, WebhookSignature  # noqa


# Sets some basic information about the running application that's sent along
# with API requests. Useful for plugin authors to identify their plugin when
# communicating with fintecture.
#
# Takes a name and optional version and plugin URL.
def set_app_info(name, partner_id=None, url=None, version=None):
    global app_info
    app_info = {
        "name": name,
        "partner_id": partner_id,
        "url": url,
        "version": version,
    }



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")
fintecture.private_key = os.environ.get("FINTECTURE_PRIVATE_KEY")


print("Looking for URL to authorize in bank with AIS...")

# provider_id = "bschesmm"  # Banco Santander
# provider_id = "bsabesbb"  # Banco Sabadell
provider_id = "cmcifrpp"  # CIC
redirect_uri = 'https://webhook.site/5940c656-7fd3-40a0-ad41-9cfba1e986f3'

resp_authorize = fintecture.AIS.authorize(
    provider_id=provider_id,
    redirect_uri=redirect_uri,
    state="1234",

    # model='redirect', # is optional

    # model='decoupled', # is required with follow parameters too
    # psu_id='9090901',
    # psu_ip_address='92.168.0.12',
)

provider_auth_url = resp_authorize['url']
print("Redirecting user to {} ...".format(
    provider_auth_url
))
print("Success: %r\n" % (resp_authorize))

print("Now you must request in your browser previous URL, login, and finish the process...")

# if we use any bank with a sandbox, you can use some test account as follows:
# example, some test account:
# username: 078000000P
# password: 123456

print("After login... the callback redirection is requested and finish with some query parameters in your browser...")

# in cases of "error" we get URL with follow parameters:
# https://webhook.site/5940c656-7fd3-40a0-ad41-9cfba1e986f3?error=provider_error&state=1234&provider_error=access_denied
# Query Strings:
#   error: provider_error
#   state: 1234
#   provider_error: access_denied

# in cases of "success" we get URL with follow parameters:
# https://webhook.site/5940c656-7fd3-40a0-ad41-9cfba1e986f3?customer_id=380202af6694dbd2d177c84028f1908b&code=8ecd9b1079a7e1f2e4efd0a100bd5521&provider=cmcifrpp&state=1234
# Query Strings:
#   customer_id: 380202af6694dbd2d177c84028f1908b
#   code: 8ecd9b1079a7e1f2e4efd0a100bd5521
#   provider: cmcifrpp
#   state: 1234

customer_id = '54d131651e54eb7b359489bb5fba6bbc'
code = '1926ad72adccf6d0a4aa5b9dc320cb9e'
state = 1234

print("We must validate received state with sent and the same provider...")

print('Received code "{0}" and customer_id "{1}" from webhook...'.format(code, customer_id))
print("Now login using an oAuth with a code")

resp_token = fintecture.AIS.oauth(code=code)

fintecture.access_token = resp_token['access_token']

print("Success: %r" % (resp_token))

accounts = fintecture.Account.search_by_customer(customer_id)

print("accounts: %r" % (accounts))



from __future__ import absolute_import, division, print_function

import calendar
import datetime
import json
import platform
import time
import uuid
import warnings

from collections import OrderedDict
from email import utils as email_utils

import fintecture

from fintecture import crypto
from fintecture import error, oauth_error, http_client, version, util, six
from fintecture.multipart_data_generator import MultipartDataGenerator
from fintecture.six.moves.urllib.parse import urlparse, urlencode, urlsplit, urlunsplit
from fintecture.fintecture_response import FintectureResponse, FintectureStreamResponse
from fintecture.constants import SIGNED_HEADER_PARAMETER_LIST


def _encode_datetime(dttime):
    if dttime.tzinfo and dttime.tzinfo.utcoffset(dttime) is not None:
        utc_timestamp = calendar.timegm(dttime.utctimetuple())
    else:
        utc_timestamp = time.mktime(dttime.timetuple())

    return int(utc_timestamp)


def _encode_nested_dict(key, data, fmt="%s[%s]"):
    d = OrderedDict()
    for subkey, subvalue in six.iteritems(data):
        d[fmt % (key, subkey)] = subvalue
    return d


def _api_encode(data):
    for key, value in six.iteritems(data):
        key = util.utf8(key)
        if value is None:
            continue
        elif hasattr(value, "fintecture_id"):
            yield (key, value.fintecture_id)
        elif isinstance(value, list) or isinstance(value, tuple):
            for i, sv in enumerate(value):
                if isinstance(sv, dict):
                    subdict = _encode_nested_dict("%s[%d]" % (key, i), sv)
                    for k, v in _api_encode(subdict):
                        yield (k, v)
                else:
                    yield ("%s[%d]" % (key, i), util.utf8(sv))
        elif isinstance(value, dict):
            subdict = _encode_nested_dict(key, value)
            for subkey, subvalue in _api_encode(subdict):
                yield (subkey, subvalue)
        elif isinstance(value, datetime.datetime):
            yield (key, _encode_datetime(value))
        else:
            yield (key, util.utf8(value))


def _build_api_url(url, query):
    scheme, netloc, path, base_query, fragment = urlsplit(url)

    if base_query:
        query = "%s&%s" % (base_query, query)

    return urlunsplit((scheme, netloc, path, query, fragment))


class APIRequestor(object):

    def __init__(
        self,
        app_id=None,
        app_secret=None,
        private_key=None,
        client=None,
        api_base=None,
        api_version=None,
        account=None,
    ):
        self.api_base = api_base or APIRequestor.api_base()

        self.app_id = app_id
        self.app_secret = app_secret
        self.private_key = private_key

        self.api_version = api_version or fintecture.api_version
        self.fintecture_account = account

        self._default_proxy = None

        from fintecture import verify_ssl_certs as verify
        from fintecture import proxy

        if client:
            self._client = client
        elif fintecture.default_http_client:
            self._client = fintecture.default_http_client
            if proxy != self._default_proxy:
                warnings.warn(
                    "fintecture.proxy was updated after sending a "
                    "request - this is a no-op. To use a different proxy, "
                    "set fintecture.default_http_client to a new client "
                    "configured with the proxy."
                )
        else:
            # If the fintecture.default_http_client has not been set by the user
            # yet, we'll set it here. This way, we aren't creating a new
            # HttpClient for every request.
            fintecture.default_http_client = http_client.new_default_http_client(
                verify_ssl_certs=verify, proxy=proxy
            )
            self._client = fintecture.default_http_client
            self._default_proxy = proxy

    @classmethod
    def api_base(cls):
        from fintecture import production_api_base, sandbox_api_base, env

        if env not in fintecture.AVAILABLE_ENVS:
            raise ValueError(
                "Defined environment value is invalid. "
                "Please check that available environment values is one of %r\n" % (fintecture.AVAILABLE_ENVS)
            )
        if env == fintecture.ENVIRONMENT_SANDBOX:
            return sandbox_api_base
        elif env == fintecture.ENVIRONMENT_PRODUCTION:
            return production_api_base

        return sandbox_api_base

    @classmethod
    def format_app_info(cls, info):
        str = info["name"]
        if info["version"]:
            str += "/%s" % (info["version"],)
        if info["url"]:
            str += " (%s)" % (info["url"],)
        return str

    def request(self, method, url, params=None, headers=None):
        rbody, rcode, rheaders, my_api_key = self.request_raw(
            method.lower(), url, params, headers, is_streaming=False
        )
        resp = self.interpret_response(rbody, rcode, rheaders)
        return resp, my_api_key

    def request_stream(self, method, url, params=None, headers=None):
        stream, rcode, rheaders, my_api_key = self.request_raw(
            method.lower(), url, params, headers, is_streaming=True
        )
        resp = self.interpret_streaming_response(stream, rcode, rheaders)
        return resp, my_api_key

    def handle_error_response(self, rbody, rcode, resp, rheaders):
        try:
            error_data = resp["errors"]
        except (KeyError, TypeError):
            raise error.APIError(
                "Invalid response object from API: %r (HTTP response code "
                "was %d)" % (rbody, rcode),
                rbody,
                rcode,
                resp,
            )

        err = None

        # OAuth errors are a JSON object where `error` is a string. In
        # contrast, in API errors, `error` is a hash with sub-keys. We use
        # this property to distinguish between OAuth and API errors.
        if isinstance(error_data, six.string_types):
            err = self.specific_oauth_error(
                rbody, rcode, resp, rheaders, error_data
            )

        if err is None:
            err = self.specific_api_error(
                rbody, rcode, resp, rheaders, error_data
            )

        raise err

    def specific_api_error(self, rbody, rcode, resp, rheaders, error_data):

        # only raise exception with the first error notified
        if isinstance(error_data, list) and len(error_data) > 0:
            error_data = error_data[0]

        util.log_info(
            "Fintecture API error received",
            error_code=error_data.get("code"),
            error_title=error_data.get("title"),
            error_message=error_data.get("message"),
        )

        # Rate limits were previously coded as 400's with code 'rate_limit'
        if rcode == 429 or (
            rcode == 400 and error_data.get("code") == "rate_limit"
        ):
            return error.RateLimitError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode in [400, 404]:
            if error_data.get("type") == "idempotency_error":
                return error.IdempotencyError(
                    error_data.get("message"), rbody, rcode, resp, rheaders
                )
            else:
                return error.InvalidRequestError(
                    error_data.get("message"),
                    error_data.get("param"),
                    error_data.get("code"),
                    rbody,
                    rcode,
                    resp,
                    rheaders,
                )
        elif rcode == 401:
            return error.AuthenticationError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode == 402:
            return error.CardError(
                error_data.get("message"),
                error_data.get("param"),
                error_data.get("code"),
                rbody,
                rcode,
                resp,
                rheaders,
            )
        elif rcode == 403:
            return error.PermissionError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        else:
            return error.APIError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )

    def specific_oauth_error(self, rbody, rcode, resp, rheaders, error_code):
        description = resp.get("error_description", error_code)

        util.log_info(
            "Fintecture OAuth error received",
            error_code=error_code,
            error_description=description,
        )

        args = [error_code, description, rbody, rcode, resp, rheaders]

        if error_code == "invalid_client":
            return oauth_error.InvalidClientError(*args)
        elif error_code == "invalid_grant":
            return oauth_error.InvalidGrantError(*args)
        elif error_code == "invalid_request":
            return oauth_error.InvalidRequestError(*args)
        elif error_code == "invalid_scope":
            return oauth_error.InvalidScopeError(*args)
        elif error_code == "unsupported_grant_type":
            return oauth_error.UnsupportedGrantTypError(*args)
        elif error_code == "unsupported_response_type":
            return oauth_error.UnsupportedResponseTypError(*args)

        return None

    def request_headers(self, app_id, method, url, access_token=None, private_key=None, body=None):
        user_agent = "Fintecture/v1 PythonBindings/%s" % (version.VERSION,)
        if fintecture.app_info:
            user_agent += " " + self.format_app_info(fintecture.app_info)

        # ua = {
        #     "bindings_version": version.VERSION,
        #     "lang": "python",
        #     "publisher": "fintecture",
        #     "httplib": self._client.name,
        # }
        # for attr, func in [
        #     ["lang_version", platform.python_version],
        #     ["platform", platform.platform],
        #     ["uname", lambda: " ".join(platform.uname())],
        # ]:
        #     try:
        #         val = func()
        #     except Exception:
        #         val = "(disabled)"
        #     ua[attr] = val
        # if fintecture.app_info:
        #     ua["application"] = fintecture.app_info

        headers = {
            # "X-Fintecture-Client-User-Agent": json.dumps(ua),
            "User-Agent": user_agent,
            "Accept": "application/json",
        }

        # if self.fintecture_account:
        #     headers["Fintecture-Account"] = self.fintecture_account

        if method == "post":
            headers["Content-Type"] = "application/x-www-form-urlencoded"
            headers.setdefault("Idempotency-Key", str(uuid.uuid4()))

        if self.api_version is not None:
            headers["Fintecture-Version"] = self.api_version

        if app_id:
            headers['app_id'] = app_id

        if access_token:
            headers['Authorization'] = "Bearer {}".format(fintecture.access_token)

        if private_key:
            payload = ''
            if type(body) == 'str':
                payload = body
            elif isinstance(body, dict):
                payload = json.dumps(body, separators=(',', ':'))

            parsed_url = urlparse(url)
            path = parsed_url.path
            query = parsed_url.query

            if payload and len(payload) > 0:
                headers['Digest'] = 'SHA-256=' + crypto.hash_base64(payload)

            headers['Date'] = email_utils.format_datetime(datetime.datetime.utcnow())
            headers['X-Date'] = headers['Date']
            headers['X-Request-ID'] = crypto.generate_uuidv4()
            headers['(request-target)'] = method + ' ' + path + ('?' + query if query else '')
            headers['Signature'] = crypto.create_signature_header(
                headers, app_id, private_key, SIGNED_HEADER_PARAMETER_LIST)
            del headers['(request-target)']

        return headers

    def request_raw(
        self,
        method,
        url,
        params=None,
        supplied_headers=None,
        is_streaming=False,
    ):
        """
        Mechanism for issuing an API call
        """

        if self.app_id:
            my_app_id = self.app_id
        else:
            from fintecture import app_id
            my_app_id = app_id

        if my_app_id is None:
            raise error.AuthenticationError(
                "No app_id and/or app_secret provided. (HINT: set your app_id and app_secret using "
                '"fintecture.app_id = <APP-ID>" and "fintecture.app_secret = <APP-SECRET>"). '
                "You can find your application values in your Fintecture developer console at "
                "https://console.fintecture.com/developers, after registering your account as a platform. "
                "See https://docs.fintecture.com/ for details, or email contact@fintecture.com "
                "if you have any questions."
            )

        if self.app_secret:
            my_app_secret = self.app_secret
        else:
            from fintecture import app_secret
            my_app_secret = app_secret

        if self.private_key:
            my_private_key = self.private_key
        else:
            from fintecture import private_key
            my_private_key = private_key

        abs_url = "%s%s" % (self.api_base, url)

        encoded_params = urlencode(list(_api_encode(params or {})))

        # Don't use strict form encoding by changing the square bracket control
        # characters back to their literals. This is fine by the server, and
        # makes these parameter strings easier to read.
        encoded_params = encoded_params.replace("%5B", "[").replace("%5D", "]")

        if method == "get" or method == "delete":
            if params:
                abs_url = _build_api_url(abs_url, encoded_params)
            post_data = None
        elif method == "post" or method == "patch":
            if (
                supplied_headers is not None
                and supplied_headers.get("Content-Type")
                == "multipart/form-data"
            ):
                generator = MultipartDataGenerator()
                generator.add_params(params or {})
                post_data = generator.get_post_data()
                supplied_headers[
                    "Content-Type"
                ] = "multipart/form-data; boundary=%s" % (generator.boundary,)
            elif (
                supplied_headers is not None
                and supplied_headers.get("Content-Type")
                == "application/json"
            ):
                post_data = json.dumps(params)
            else:
                post_data = encoded_params
        else:
            raise error.APIConnectionError(
                "Unrecognized HTTP method %r.  This may indicate a bug in the "
                "Fintecture bindings.  Please contact contact@fintecture.com for "
                "assistance." % (method,)
            )

        headers = self.request_headers(my_app_id, method, abs_url, fintecture.access_token, my_private_key, post_data)
        if supplied_headers is not None:
            for key, value in six.iteritems(supplied_headers):
                headers[key] = value

        util.log_info("Request to Fintecture api", method=method, path=abs_url)
        util.log_debug(
            "Post details",
            post_data=encoded_params,
            api_version=self.api_version,
        )

        if is_streaming:
            (
                rcontent,
                rcode,
                rheaders,
            ) = self._client.request_stream_with_retries(
                method, abs_url, headers, post_data
            )
        else:
            rcontent, rcode, rheaders = self._client.request_with_retries(
                method, abs_url, headers, post_data
            )

        util.log_info("Fintecture API response", path=abs_url, response_code=rcode)
        util.log_debug("API response body", body=rcontent)

        if "Request-Id" in rheaders:
            request_id = rheaders["Request-Id"]
            util.log_debug(
                "Dashboard link for request",
                link=util.dashboard_link(request_id),
            )

        return rcontent, rcode, rheaders, my_app_id

    def _should_handle_code_as_error(self, rcode):
        return not 200 <= rcode < 300

    def interpret_response(self, rbody, rcode, rheaders):
        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")
            resp = FintectureResponse(rbody, rcode, rheaders)
        except Exception:
            raise error.APIError(
                "Invalid response body from API: %s "
                "(HTTP response code was %d)" % (rbody, rcode),
                rbody,
                rcode,
                rheaders,
            )
        if self._should_handle_code_as_error(rcode):
            self.handle_error_response(rbody, rcode, resp.data, rheaders)
        return resp

    def interpret_streaming_response(self, stream, rcode, rheaders):
        # Streaming response are handled with minimal processing for the success
        # case (ie. we don't want to read the content). When an error is
        # received, we need to read from the stream and parse the received JSON,
        # treating it like a standard JSON response.
        if self._should_handle_code_as_error(rcode):
            if hasattr(stream, "getvalue"):
                json_content = stream.getvalue()
            elif hasattr(stream, "read"):
                json_content = stream.read()
            else:
                raise NotImplementedError(
                    "HTTP client %s does not return an IOBase object which "
                    "can be consumed when streaming a response."
                )

            return self.interpret_response(json_content, rcode, rheaders)
        else:
            return FintectureStreamResponse(stream, rcode, rheaders)



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")
fintecture.private_key = os.environ.get("FINTECTURE_PRIVATE_KEY")
fintecture.access_token = os.environ.get("FINTECTURE_ACCESS_TOKEN")


print("Requesting to pay...")

if fintecture.access_token is None:
    print("Requesting an access token...")

    resp = fintecture.PIS.oauth()

    print("Success: %r" % (resp))

    fintecture.access_token = resp['access_token']
else:
    print("Using access token: %s" % fintecture.access_token)

redirect_uri = "https://domain.com"
print("Doing a request for payout call with redirection URL to %s..." % redirect_uri)

resp_request_for_payout = fintecture.PIS.request_to_payout(
    redirect_uri,
    language='es',
    state='85321',
    meta={
        'psu_name': 'Jean',
        'psu_email': 'jean@hooker.com',
        'psu_phone': '0232420304',
        'psu_phone_prefix': '+33',
        'psu_address': {
            'street_number': '8',
            'street': 'Alan Parvis Turing',
            'zip': '75013',
            'city': 'Paris',
            'country': 'FR',
        },
        'expire': 86400,
        'method': 'link',
    },
    data={
        'attributes': {
            'amount': '101.97',
            'currency': 'EUR',
            'communication': 'Reference 6469878'
        }
    }
)

print("Success: %r" % (resp_request_for_payout))

print("Doing a request to pay call with redirection URL to %s..." % redirect_uri)

resp_request_to_pay = fintecture.PIS.request_to_pay(
    redirect_uri,
    language='es',
    meta={
        'psu_name': 'M. John Doe',
        'psu_email': 'john@doe.com',
        'psu_phone': '0601020304',
        'psu_phone_prefix': '+33',
        'psu_address': {
            'street_number': '5',
            'street': 'Parvis Alan Turing',
            'zip': '75013',
            'city': 'Paris',
            'country': 'FR',
        },
        'expire': 86400,
        'due_date': 76400,
        'cc': 'copyc@mail.com',
        'bcc': 'tri_copyc@mail.com',
    },
    data={
        'type': 'request-to-pay',
        'attributes': {
            'amount': '230.79',
            'currency': 'EUR',
            'communication': 'Reference 8786469'
        }
    }
)

print("Success: %r" % (resp_request_to_pay))

# We are receiving an invalid response with HTTP with 500 of status code:
#
# fintecture.error.APIError: Invalid response object from API:
# '{
#   "status":"500",
#   "code":"internal_server_error",
#   "log_id":"b9bef7a8-0981-46da-8317-f0d1d2b7984b",
#   "errors":[
#       {
#           "code":"request_to_pay_unsuccessful",
#           "title":"Internal Error",
#           "detail":"The request to pay initiation has been unsuccessful"
#       }
#   ]
# }'
# (HTTP response code was 500)



# File generated from our OpenAPI spec
from __future__ import absolute_import, division, print_function

from fintecture.api_resources.abstract import CreateableAPIResource
from fintecture.api_resources.abstract import DeletableAPIResource
from fintecture.api_resources.abstract import ListableAPIResource
from fintecture.api_resources.abstract import UpdateableAPIResource


class Account(
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
):
    OBJECT_NAME = "account"

    @classmethod
    def search_by_customer(cls, customer_id, **params):
        return cls._static_request(
            "get",
            "/ais/v1/customer/{}/accounts".format(
                customer_id
            ),
            params=params,
        )

    @classmethod
    def search_transactions_by_customer_account(cls, customer_id, account_id, **params):
        return cls._static_request(
            "get",
            "/ais/v1/customer/{}/accounts/{}/transactions".format(
                customer_id,
                account_id
            ),
            params=params,
        )

    @classmethod
    def delete(cls, customer_id, **params):
        return cls._cls_delete(customer_id, **params)

    @classmethod
    def search(cls, *args, **kwargs):
        return cls._search(search_url="/res/v1/providers", *args, **kwargs)

    @classmethod
    def search_auto_paging_iter(cls, *args, **kwargs):
        return cls.search(*args, **kwargs).auto_paging_iter()



from __future__ import absolute_import, division, print_function

import os

import fintecture


fintecture.app_id = os.environ.get("FINTECTURE_APP_ID")
fintecture.app_secret = os.environ.get("FINTECTURE_APP_SECRET")
fintecture.private_key = os.environ.get("FINTECTURE_PRIVATE_KEY")
fintecture.access_token = os.environ.get("FINTECTURE_ACCESS_TOKEN")


print("Refunding an initiate payment...")

if fintecture.access_token is None:
    print("Requesting an access token...")

    resp = fintecture.PIS.oauth()

    print("Success: %r" % (resp))

    fintecture.access_token = resp['access_token']
else:
    print("Using access token: %s" % fintecture.access_token)

session_id = '88a6c01995a642a1924c465503f33cc4'

payment = fintecture.Payment.retrieve(session_id)

print("Retrieving refund for selected session...")

resp_session_payments_refund = payment.refund(data={
    'attributes': {
        "amount": "15.2",
        "communication": "REFUND ORDER-123"
    }
})

print("Success: %r" % (resp_session_payments_refund))

resp_initiate_refund = fintecture.PIS.initiate_refund(
    state='5678',
    meta={
        'session_id': session_id
    },
    data={
        'attributes': {
            'amount': '1.99',
            'communication': 'REFUND ORDER-123'
        }
    })

print("Success: %r" % (resp_initiate_refund))




ENVIRONMENT_TEST = 'test'
ENVIRONMENT_SANDBOX = 'sandbox'
ENVIRONMENT_PRODUCTION = 'production'

DEFAULT_ENV = ENVIRONMENT_SANDBOX

AVAILABLE_ENVS = [
    ENVIRONMENT_TEST,
    ENVIRONMENT_SANDBOX,
    ENVIRONMENT_PRODUCTION,
]
